import pandas as pd
import pandas_ta as ta
import uvicorn
import traceback
import requests
import time

import warnings

from fastapi import FastAPI, HTTPException
from typing import Optional, List
from pydantic import BaseModel
from datetime import datetime

# Add these lines at the beginning of your script
warnings.filterwarnings('ignore', category=FutureWarning)

pd.set_option('display.max_rows', None)

app = FastAPI(title="Crypto Data Service", description="API for fetching and analyzing crypto data")

class IndicatorResponse(BaseModel):
    symbol: str
    interval: str
    kline_time: int
    indicator_time: int
    proxy_server: str
    data: List

def process_kline_data(symbol, interval, limit, from_timestamp=None):
    """Process raw kline data into a pandas DataFrame"""

    start_ms = int(round(time.time() * 1000))

    url = f'http://127.0.0.1/indicators/api/klines/?symbol={symbol}&interval={interval}&limit={limit}&from_timestamp={from_timestamp}'
    response = requests.get(url)

    # X-Proxy-Server header
    proxy_server = response.headers.get('X-Proxy-Server', None)

    kline_data = []
    if response.status_code == 200:
        kline_data = response.json()

    end_ms = int(round(time.time() * 1000))
    kline_time = end_ms - start_ms

    # Check kline data not empty
    if len(kline_data) == 0:
        raise HTTPException(
            status_code=500,
            detail=f"Failed to fetch kline data for {symbol} at {interval} interval with limit {limit} and process time {kline_time} ms and proxy server {proxy_server}"
        )

    data = [
        [
            entry[0],  # timestamp
            entry[1],  # open
            entry[2],  # high
            entry[3],  # low
            entry[4],  # close
            entry[5],  # volume
        ]
        for entry in kline_data
    ]

    df = pd.DataFrame(data, columns=["timestamp", "open", "high", "low", "close", "volume"])

    # Convert timestamp to datetime
    df['timestamp'] = pd.to_datetime(df['timestamp'], unit='ms')

    for col in ["open", "high", "low", "close", "volume"]:
        df[col] = df[col].astype(float)

    df.set_index(pd.DatetimeIndex(df["timestamp"]), inplace=True)

    # return df, kline_time, proxy_server
    return df, kline_time, proxy_server

"""Relative Strength Index (RSI)"""
@app.get("/rsi/{symbol}")
async def get_rsi(
    symbol: str,
    interval: str = '1h',
    limit: Optional[int] = 2000,
    from_timestamp: Optional[int] = None,
):
    try:
        # Fetch kline data
        symbol = symbol.upper()

        # Process data
        df, kline_time, proxy_server = process_kline_data(symbol, interval, limit + 100, from_timestamp)

        start_ms = int(round(time.time() * 1000))

        # Calculate RSI
        rsi = df.ta.rsi(length=14)

        # Convert RSI to a DataFrame for consistent response formatting
        rsi_df = rsi.reset_index()
        rsi_df.columns = ['timestamp', 'value']

        # Convert the timestamp to seconds since the Unix epoch
        rsi_df['timestamp'] = pd.to_datetime(rsi_df['timestamp']).astype(int) // 10**9

        # Replace NaN and infinite values with 0
        rsi_df.fillna(value=0, inplace=True)

        # Trim the DataFrame to the specified limit
#         rsi_df = rsi_df.tail(limit)

        end_ms = int(round(time.time() * 1000))
        indicator_time = end_ms - start_ms

        return {
            "symbol": symbol,
            "interval": interval,
            "kline_time": kline_time,
            "indicator_time": indicator_time,
            "proxy_server": proxy_server,
            "data": rsi_df.to_dict(orient='records')
        }
    except Exception as e:
        print(e)
        print(traceback.format_exc())
        raise HTTPException(
            status_code=500,
            detail=f"{str(e)}"
        )

"""Exponential Moving Average (EMA)"""
@app.get("/ema/{symbol}")
async def get_ema(
    symbol: str,
    interval: str = '1h',
    limit: Optional[int] = 2000,
    from_timestamp: Optional[int] = None,
):
    try:
        # Fetch kline data
        symbol = symbol.upper()

        # Process data
        df, kline_time, proxy_server = process_kline_data(symbol, interval, limit + 100, from_timestamp)

        start_ms = int(round(time.time() * 1000))

        # Calculate
        ema = df.ta.ema(length=20)

        # Convert RSI to a DataFrame for consistent response formatting
        ema_df = ema.reset_index()
        ema_df.columns = ['timestamp', 'value']

        # Convert the timestamp to seconds since the Unix epoch
        ema_df['timestamp'] = pd.to_datetime(ema_df['timestamp']).astype(int) // 10**9

        # Replace NaN and infinite values with 0
        ema_df.fillna(value=0, inplace=True)

        # Trim the DataFrame to the specified limit
#         ema_df = ema_df.tail(limit)

        end_ms = int(round(time.time() * 1000))
        indicator_time = end_ms - start_ms

        return {
            "symbol": symbol,
            "interval": interval,
            "kline_time": kline_time,
            "indicator_time": indicator_time,
            "proxy_server": proxy_server,
            "data": ema_df.to_dict(orient='records')
        }
    except Exception as e:
        print(e)
        print(traceback.format_exc())
        raise HTTPException(
            status_code=500,
            detail=f"{str(e)}"
        )

"""Chaikin Money Flow (CMF)"""
@app.get("/cmf/{symbol}")
async def get_cmf(
    symbol: str,
    interval: str = '1h',
    limit: Optional[int] = 2000,
    from_timestamp: Optional[int] = None,
):
    try:
        # Fetch kline data
        symbol = symbol.upper()

        # Process data
        df, kline_time, proxy_server = process_kline_data(symbol, interval, limit + 100, from_timestamp)

        start_ms = int(round(time.time() * 1000))

        # Calculate
        cmf = df.ta.cmf(length=20)

        # Convert RSI to a DataFrame for consistent response formatting
        cmf_df = cmf.reset_index()
        cmf_df.columns = ['timestamp', 'value']

        # Convert the timestamp to seconds since the Unix epoch
        cmf_df['timestamp'] = pd.to_datetime(cmf_df['timestamp']).astype(int) // 10**9

        # Replace NaN and infinite values with 0
        cmf_df.fillna(value=0, inplace=True)

        # Trim the DataFrame to the specified limit
#         cmf_df = cmf_df.tail(limit)

        end_ms = int(round(time.time() * 1000))
        indicator_time = end_ms - start_ms

        return {
            "symbol": symbol,
            "interval": interval,
            "kline_time": kline_time,
            "indicator_time": indicator_time,
            "proxy_server": proxy_server,
            "data": cmf_df.to_dict(orient='records')
        }
    except Exception as e:
        print(e)
        print(traceback.format_exc())
        raise HTTPException(
            status_code=500,
            detail=f"{str(e)}"
        )

"""Money Flow Index (MFI)"""
@app.get("/mfi/{symbol}")
async def get_mfi(
    symbol: str,
    interval: str = '1h',
    limit: Optional[int] = 2000,
    from_timestamp: Optional[int] = None,
):
    try:
        # Fetch kline data
        symbol = symbol.upper()

        # Process data
        df, kline_time, proxy_server = process_kline_data(symbol, interval, limit + 100, from_timestamp)

        start_ms = int(round(time.time() * 1000))

        length = 14

        # Drop duplicate timestamps if they exist
        df = df.drop_duplicates(subset=['timestamp'], keep='first')

        # length based on number of df and max is 14
        if len(df) < 14:
            length = len(df)

        # Set timestamp as index after removing duplicates
        df = df.set_index('timestamp')

        # Calculate
        mfi = df.ta.mfi(length=length)

        # Convert RSI to a DataFrame for consistent response formatting
        mfi_df = mfi.reset_index()
        mfi_df.columns = ['timestamp', 'value']

        mfi_df['timestamp'] = pd.to_datetime(mfi_df['timestamp']).astype(int) // 10**9

        # Replace NaN and infinite values with 0
        mfi_df.fillna(value=0, inplace=True)

        # Trim the DataFrame to the specified limit
#         mfi_df = mfi_df.tail(limit)

        end_ms = int(round(time.time() * 1000))
        indicator_time = end_ms - start_ms

        return {
            "symbol": symbol,
            "interval": interval,
            "kline_time": kline_time,
            "indicator_time": indicator_time,
            "proxy_server": proxy_server,
            "data": mfi_df.to_dict(orient='records')
        }
    except Exception as e:
        print(e)
        print(traceback.format_exc())
        raise HTTPException(
            status_code=500,
            detail=f"{str(e)}"
        )

"""Lowest and highest values over a specified period (minmax)"""
@app.get("/minmax/{symbol}")
async def get_minmax(
    symbol: str,
    interval: str = '1h',
    limit: Optional[int] = 2000,
    from_timestamp: Optional[int] = None,
):
    try:
        # Fetch kline data
        symbol = symbol.upper()

        # Process data
        df, kline_time, proxy_server = process_kline_data(symbol, interval, limit + 100, from_timestamp)

        start_ms = int(round(time.time() * 1000))

        # Specify the period
        period = 30

        # Calculate the lowest and highest values over the period
        df["lowest"] = df["close"].rolling(window=period).min()
        df["highest"] = df["close"].rolling(window=period).max()

        # Add Lowest and Highest as Custom Indicator
        df.ta.strategy({
            "name": "Lowest",
            "function": lambda close, length: close.rolling(window=length).min(),
            "params": {"length": period},
        })

        df.ta.strategy({
            "name": "Highest",
            "function": lambda close, length: close.rolling(window=length).max(),
            "params": {"length": period},
        })

        # Create minmax DataFrame
        df.fillna(value=0, inplace=True)

        # convert timestamp to seconds since the Unix epoch
        df['timestamp'] = pd.to_datetime(df['timestamp']).astype(int) // 10**9

        # Remove open, high, low, close, volume columns
        df.drop(columns=['open', 'high', 'low', 'close', 'volume'], inplace=True)

        # Trim the DataFrame to the specified limit
#         df = df.tail(limit)

        end_ms = int(round(time.time() * 1000))
        indicator_time = end_ms - start_ms

        return {
            "symbol": symbol,
            "interval": interval,
            "kline_time": kline_time,
            "indicator_time": indicator_time,
            "proxy_server": proxy_server,
            "data": df.to_dict(orient='records')
        }
    except Exception as e:
        print(e)
        print(traceback.format_exc())
        raise HTTPException(
            status_code=500,
            detail=f"{str(e)}"
        )

"""Moving Average (MA)"""
@app.get("/ma/{symbol}")
async def get_ma(
    symbol: str,
    interval: str = '1h',
    limit: Optional[int] = 2000,
    from_timestamp: Optional[int] = None,
):
    try:
        # Fetch kline data
        symbol = symbol.upper()

        # Process data
        df, kline_time, proxy_server = process_kline_data(symbol, interval, limit + 100, from_timestamp)

        start_ms = int(round(time.time() * 1000))

        # Specify the length
        length = 20

        df["value"] = df["close"].rolling(window=length).mean()

        # Add MA as Custom Indicator
        df.ta.strategy({
            "name": "MA",
            "function": lambda close, length: close.rolling(window=length).mean(),
            "params": {"length": length},
        })

        # Create MA DataFrame
        ma_df = df[["timestamp", "value"]].copy()
        ma_df.fillna(value=0, inplace=True)

        # convert timestamp to seconds since the Unix epoch
        ma_df['timestamp'] = pd.to_datetime(ma_df['timestamp']).astype(int) // 10**9

        # limit the DataFrame to the specified limit
#         ma_df = ma_df.tail(limit)

        end_ms = int(round(time.time() * 1000))
        indicator_time = end_ms - start_ms

        return {
            "symbol": symbol,
            "interval": interval,
            "kline_time": kline_time,
            "indicator_time": indicator_time,
            "proxy_server": proxy_server,
            "data": ma_df.to_dict(orient='records')
        }
    except Exception as e:
        print(e)
        print(traceback.format_exc())
        raise HTTPException(
            status_code=500,
            detail=f"{str(e)}"
        )

"""Rate of Change (ROC)"""
@app.get("/roc/{symbol}")
async def get_roc(
    symbol: str,
    interval: str = '1h',
    limit: Optional[int] = 2000,
    from_timestamp: Optional[int] = None,
):
    try:
        # Fetch kline data
        symbol = symbol.upper()

        # Process data
        df, kline_time, proxy_server = process_kline_data(symbol, interval, limit + 100, from_timestamp)

        start_ms = int(round(time.time() * 1000))

        # Specify the period
        length = 10

        # Calculate
        indicator = df.ta.roc(length=length)

        # Convert RSI to a DataFrame for consistent response formatting
        indicator_df = indicator.reset_index()
        indicator_df.columns = ['timestamp', 'value']

        # Convert the timestamp to seconds since the Unix epoch
        indicator_df['timestamp'] = pd.to_datetime(indicator_df['timestamp']).astype(int) // 10**9

        # Replace NaN and infinite values with 0
        indicator_df.fillna(value=0, inplace=True)

        # limit the DataFrame to the specified limit
#         indicator_df = indicator_df.tail(limit)

        end_ms = int(round(time.time() * 1000))
        indicator_time = end_ms - start_ms

        return {
            "symbol": symbol,
            "interval": interval,
            "kline_time": kline_time,
            "indicator_time": indicator_time,
            "proxy_server": proxy_server,
            "data": indicator_df.to_dict(orient='records')
        }
    except Exception as e:
        print(e)
        print(traceback.format_exc())
        raise HTTPException(
            status_code=500,
            detail=f"{str(e)}"
        )

"""Parabolic SAR (PSAR)"""
@app.get("/psar/{symbol}")
async def get_psar(
    symbol: str,
    interval: str = '1h',
    limit: Optional[int] = 2000,
    from_timestamp: Optional[int] = None,
):
    try:
        # Fetch kline data
        symbol = symbol.upper()

        # Process data
        df, kline_time, proxy_server = process_kline_data(symbol, interval, limit + 100, from_timestamp)

        start_ms = int(round(time.time() * 1000))

        # Calculate
        indicator = df.ta.psar(high=df["high"], low=df["low"], close=df["close"], af0=0.02, af=0.02, max_af=0.2)

        # Convert RSI to a DataFrame for consistent response formatting
        indicator_df = indicator.reset_index()
        indicator_df.columns = ['timestamp', '1', 'value', '2', '3']

        # Keep only the timestamp and value columns
        indicator_df = indicator_df[['timestamp', 'value']]

        # Convert the timestamp to seconds since the Unix epoch
        indicator_df['timestamp'] = pd.to_datetime(indicator_df['timestamp']).astype(int) // 10**9

        # Replace NaN and infinite values with 0
        indicator_df.fillna(value=0, inplace=True)

        # limit the DataFrame to the specified limit
#         indicator_df = indicator_df.tail(limit)

        end_ms = int(round(time.time() * 1000))
        indicator_time = end_ms - start_ms

        return {
            "symbol": symbol,
            "interval": interval,
            "kline_time": kline_time,
            "indicator_time": indicator_time,
            "proxy_server": proxy_server,
            "data": indicator_df.to_dict(orient='records')
        }
    except Exception as e:
        print(e)
        print(traceback.format_exc())
        raise HTTPException(
            status_code=500,
            detail=f"{str(e)}"
        )

"""Super Trend"""
@app.get("/supertrend/{symbol}")
async def supertrend(
    symbol: str,
    interval: str = '1h',
    limit: Optional[int] = 2000,
    from_timestamp: Optional[int] = None,
):
    try:
        # Fetch kline data
        symbol = symbol.upper()

        # Process data
        df, kline_time, proxy_server = process_kline_data(symbol, interval, limit + 500, from_timestamp)

        start_ms = int(round(time.time() * 1000))

        # Parameters for Supertrend
        length = 20  # Period for ATR calculation
        multiplier = 3  # Multiplier for ATR

        # Calculate the Supertrend
        indicator = df.ta.supertrend(length=length, multiplier=multiplier)
        indicator_df = indicator.reset_index()
        indicator_df.columns = ['timestamp', 'value', 'direction', 'lower', 'upper']

        # Convert the timestamp to seconds since the Unix epoch
        indicator_df['timestamp'] = pd.to_datetime(indicator_df['timestamp']).astype(int) // 10**9

        # Keep only the timestamp, value and direction columns
        indicator_df = indicator_df[['timestamp', 'value', 'direction']]

        # Replace NaN and infinite values with 0
        indicator_df.fillna(value=0, inplace=True)

        # limit the DataFrame to the specified limit
#         indicator_df = indicator_df.tail(limit)

        end_ms = int(round(time.time() * 1000))
        indicator_time = end_ms - start_ms

        return {
            "symbol": symbol,
            "interval": interval,
            "kline_time": kline_time,
            "indicator_time": indicator_time,
            "proxy_server": proxy_server,
            "data": indicator_df.to_dict(orient='records')
        }
    except Exception as e:
        print(e)
        print(traceback.format_exc())
        raise HTTPException(
            status_code=500,
            detail=f"{str(e)}"
        )

"""Volume Weighted Average Price (VWAP)"""
@app.get("/vwap/{symbol}")
async def get_vwap(
    symbol: str,
    interval: str = '1h',
    limit: Optional[int] = 2000,
    from_timestamp: Optional[int] = None,
):
    try:
        # Fetch kline data
        symbol = symbol.upper()

        # Process data
        df, kline_time, proxy_server = process_kline_data(symbol, interval, limit + 500, from_timestamp)

        start_ms = int(round(time.time() * 1000))

        # Parameters for vwap
        period = 20

        # Calculate the VWAP
        indicator = df.ta.vwap()
        indicator_df = indicator.reset_index()
        indicator_df.columns = ['timestamp', 'value']

        # Convert the timestamp to seconds since the Unix epoch
        indicator_df['timestamp'] = pd.to_datetime(indicator_df['timestamp']).astype(int) // 10**9

        # Replace NaN and infinite values with 0
        indicator_df.fillna(value=0, inplace=True)

        # limit the DataFrame to the specified limit
#         indicator_df = indicator_df.tail(limit)

        end_ms = int(round(time.time() * 1000))
        indicator_time = end_ms - start_ms

        return {
            "symbol": symbol,
            "interval": interval,
            "kline_time": kline_time,
            "indicator_time": indicator_time,
            "proxy_server": proxy_server,
            "data": indicator_df.to_dict(orient='records')
        }
    except Exception as e:
        print(e)
        print(traceback.format_exc())
        raise HTTPException(
            status_code=500,
            detail=f"{str(e)}"
        )

"""Lowest value over a specified period (min)"""
@app.get("/min/{symbol}")
async def get_min(
    symbol: str,
    interval: str = '1h',
    limit: Optional[int] = 2000,
    from_timestamp: Optional[int] = None,
):
    try:
        # Fetch kline data
        symbol = symbol.upper()

        # Process data
        df, kline_time, proxy_server = process_kline_data(symbol, interval, limit + 100, from_timestamp)

        start_ms = int(round(time.time() * 1000))

        # Parameters for vwap
        period = 20

        # Find the lowest value over the period using only pandas
        indicator = df['close'].rolling(window=period).min()

#         # Calculate the min
        indicator_df = indicator.reset_index()
        indicator_df.columns = ['timestamp', 'value']

        # Convert the timestamp to seconds since the Unix epoch
        indicator_df['timestamp'] = pd.to_datetime(indicator_df['timestamp']).astype(int) // 10**9

        # Replace NaN and infinite values with 0
        indicator_df.fillna(value=0, inplace=True)

        # limit the DataFrame to the specified limit
#         indicator_df = indicator_df.tail(limit)

        end_ms = int(round(time.time() * 1000))
        indicator_time = end_ms - start_ms

        return {
            "symbol": symbol,
            "interval": interval,
            "kline_time": kline_time,
            "indicator_time": indicator_time,
            "proxy_server": proxy_server,
            "data": indicator_df.to_dict(orient='records')
        }
    except Exception as e:
        print(e)
        print(traceback.format_exc())
        raise HTTPException(
            status_code=500,
            detail=f"{str(e)}"
        )

"""TRIX"""
@app.get("/trix/{symbol}")
async def get_trix(
    symbol: str,
    interval: str = '1h',
    limit: Optional[int] = 2000,
    from_timestamp: Optional[int] = None,
):
    try:
        # Fetch kline data
        symbol = symbol.upper()

        # Process data
        df, kline_time, proxy_server = process_kline_data(symbol, interval, limit + 100, from_timestamp)

        start_ms = int(round(time.time() * 1000))

        # Parameters for vwap
        period = 30

        # Calculate the TRIX
        indicator = df.ta.trix(length=period)

        # Calculate the min
        indicator_df = indicator.reset_index()
        indicator_df.columns = ['timestamp', 'value', 'line']

        # Convert the timestamp to seconds since the Unix epoch
        indicator_df['timestamp'] = pd.to_datetime(indicator_df['timestamp']).astype(int) // 10**9

        # Replace NaN and infinite values with 0
        indicator_df.fillna(value=0, inplace=True)

        # limit the DataFrame to the specified limit
#         indicator_df = indicator_df.tail(limit)

        end_ms = int(round(time.time() * 1000))
        indicator_time = end_ms - start_ms

        return {
            "symbol": symbol,
            "interval": interval,
            "kline_time": kline_time,
            "indicator_time": indicator_time,
            "proxy_server": proxy_server,
            "data": indicator_df.to_dict(orient='records')
        }
    except Exception as e:
        print(e)
        print(traceback.format_exc())
        raise HTTPException(
            status_code=500,
            detail=f"{str(e)}"
        )

"""Balance Of Power (BOP)"""
@app.get("/bop/{symbol}")
async def get_bop(
    symbol: str,
    interval: str = '1h',
    limit: Optional[int] = 2000,
    from_timestamp: Optional[int] = None,
):
    try:
        # Fetch kline data
        symbol = symbol.upper()

        # Process data
        df, kline_time, proxy_server = process_kline_data(symbol, interval, limit + 100, from_timestamp)

        start_ms = int(round(time.time() * 1000))

        # Calculate the TRIX
        indicator = df.ta.bop()

        # Calculate the min
        indicator_df = indicator.reset_index()
        indicator_df.columns = ['timestamp', 'value']

        # Convert the timestamp to seconds since the Unix epoch
        indicator_df['timestamp'] = pd.to_datetime(indicator_df['timestamp']).astype(int) // 10**9

        # Replace NaN and infinite values with 0
        indicator_df.fillna(value=0, inplace=True)

        # limit the DataFrame to the specified limit
#         indicator_df = indicator_df.tail(limit)

        end_ms = int(round(time.time() * 1000))
        indicator_time = end_ms - start_ms

        return {
            "symbol": symbol,
            "interval": interval,
            "kline_time": kline_time,
            "indicator_time": indicator_time,
            "proxy_server": proxy_server,
            "data": indicator_df.to_dict(orient='records')
        }
    except Exception as e:
        print(e)
        print(traceback.format_exc())
        raise HTTPException(
            status_code=500,
            detail=f"{str(e)}"
        )

"""Percentage Price Oscillator (PPO)"""
@app.get("/ppo/{symbol}")
async def get_ppo(
    symbol: str,
    interval: str = '1h',
    limit: Optional[int] = 2000,
    from_timestamp: Optional[int] = None,
):
    try:
        # Fetch kline data
        symbol = symbol.upper()

        # Process data
        df, kline_time, proxy_server = process_kline_data(symbol, interval, limit + 100, from_timestamp)

        start_ms = int(round(time.time() * 1000))

        # Calculate the TRIX
        indicator = df.ta.ppo()

        # Calculate the min
        indicator_df = indicator.reset_index()
        indicator_df.columns = ['timestamp', 'value', 'histogram', 'signal']

        # Convert the timestamp to seconds since the Unix epoch
        indicator_df['timestamp'] = pd.to_datetime(indicator_df['timestamp']).astype(int) // 10**9

        # Replace NaN and infinite values with 0
        indicator_df.fillna(value=0, inplace=True)

        # limit the DataFrame to the specified limit
#         indicator_df = indicator_df.tail(limit)

        end_ms = int(round(time.time() * 1000))
        indicator_time = end_ms - start_ms

        return {
            "symbol": symbol,
            "interval": interval,
            "kline_time": kline_time,
            "indicator_time": indicator_time,
            "proxy_server": proxy_server,
            "data": indicator_df.to_dict(orient='records')
        }
    except Exception as e:
        print(e)
        print(traceback.format_exc())
        raise HTTPException(
            status_code=500,
            detail=f"{str(e)}"
        )

"""Vortex"""
@app.get("/vortex/{symbol}")
async def get_vortex(
    symbol: str,
    interval: str = '1h',
    limit: Optional[int] = 2000,
    from_timestamp: Optional[int] = None,
):
    try:
        # Fetch kline data
        symbol = symbol.upper()

        # Process data
        df, kline_time, proxy_server = process_kline_data(symbol, interval, limit + 100, from_timestamp)

        start_ms = int(round(time.time() * 1000))

        # Calculate the Vortex
        indicator = df.ta.vortex()

        # Calculate the min
        indicator_df = indicator.reset_index()
        indicator_df.columns = ['timestamp', 'plus', 'minus']

        # Convert the timestamp to seconds since the Unix epoch
        indicator_df['timestamp'] = pd.to_datetime(indicator_df['timestamp']).astype(int) // 10**9

        # Replace NaN and infinite values with 0
        indicator_df.fillna(value=0, inplace=True)

        # limit the DataFrame to the specified limit
#         indicator_df = indicator_df.tail(limit)

        end_ms = int(round(time.time() * 1000))
        indicator_time = end_ms - start_ms

        return {
            "symbol": symbol,
            "interval": interval,
            "kline_time": kline_time,
            "indicator_time": indicator_time,
            "proxy_server": proxy_server,
            "data": indicator_df.to_dict(orient='records')
        }
    except Exception as e:
        print(e)
        print(traceback.format_exc())
        raise HTTPException(
            status_code=500,
            detail=f"{str(e)}"
        )

"""TD Sequential"""
@app.get("/tdsequential/{symbol}")
async def get_tdsequential(
    symbol: str,
    interval: str = '1h',
    limit: Optional[int] = 2000,
    from_timestamp: Optional[int] = None,
):
    try:
        # Fetch kline data
        symbol = symbol.upper()

        # Process data
        df, kline_time, proxy_server = process_kline_data(symbol, interval, limit + 100, from_timestamp)

        start_ms = int(round(time.time() * 1000))

        # Calculate RSI
        td_seq_df = df.ta.td_seq()
#         td_seq_df = td_seq.reset_index()

        # Add timestamp column
        td_seq_df['timestamp'] = df.index

#         td_seq_df.columns = ['timestamp', 'td_seq']
#         td_seq_df.columns = ['timestamp', 'rsi']

        # Convert the timestamp to seconds since the Unix epoch
        td_seq_df['timestamp'] = pd.to_datetime(td_seq_df['timestamp']).astype(int) // 10**9

        # Replace NaN and infinite values with 0
        td_seq_df.fillna(value=0, inplace=True)

        # limit the DataFrame to the specified limit
#         td_seq_df = td_seq_df.tail(limit)

        end_ms = int(round(time.time() * 1000))
        indicator_time = end_ms - start_ms

        return {
            "symbol": symbol,
            "interval": interval,
            "kline_time": kline_time,
            "indicator_time": indicator_time,
            "proxy_server": proxy_server,
            "data": td_seq_df.to_dict(orient='records')
        }
    except Exception as e:
        print(e)
        print(traceback.format_exc())
        raise HTTPException(
            status_code=500,
            detail=f"{str(e)}"
        )

"""Commodity Channel Index (CCI)"""
@app.get("/cci/{symbol}")
async def get_cci(
    symbol: str,
    interval: str = '1h',
    limit: Optional[int] = 2000,
    from_timestamp: Optional[int] = None,
):
    try:
        # Fetch kline data
        symbol = symbol.upper()

        # Process data
        df, kline_time, proxy_server = process_kline_data(symbol, interval, limit + 100, from_timestamp)

        start_ms = int(round(time.time() * 1000))

        # Calculate the Commodity Channel Index
        indicator = df.ta.cci()

        # Calculate the min
        indicator_df = indicator.reset_index()
        indicator_df.columns = ['timestamp', 'value']

        # Convert the timestamp to seconds since the Unix epoch
        indicator_df['timestamp'] = pd.to_datetime(indicator_df['timestamp']).astype(int) // 10**9

        # Replace NaN and infinite values with 0
        indicator_df.fillna(value=0, inplace=True)

        # limit the DataFrame to the specified limit
#         indicator_df = indicator_df.tail(limit)

        end_ms = int(round(time.time() * 1000))
        indicator_time = end_ms - start_ms

        return {
            "symbol": symbol,
            "interval": interval,
            "kline_time": kline_time,
            "indicator_time": indicator_time,
            "proxy_server": proxy_server,
            "data": indicator_df.to_dict(orient='records')
        }
    except Exception as e:
        print(e)
        print(traceback.format_exc())
        raise HTTPException(
            status_code=500,
            detail=f"{str(e)}"
        )

"""Bollinger Band (BBAND)"""
@app.get("/bband/{symbol}")
async def get_bband(
    symbol: str,
    interval: str = '1h',
    limit: Optional[int] = 2000,
    from_timestamp: Optional[int] = None,
):
    try:
        # Fetch kline data
        symbol = symbol.upper()

        # Process data
        df, kline_time, proxy_server = process_kline_data(symbol, interval, limit + 100, from_timestamp)

        start_ms = int(round(time.time() * 1000))

        # Calculate the Bollinger Band
        indicator = df.ta.bbands(length=20)

        # Calculate the min
        indicator_df = indicator.reset_index()
        indicator_df.columns = [
            'timestamp',
            'lower_band',
            'middle_band',
            'upper_band',
            'bandwidth',
            '%b'
        ]

        # Convert the timestamp to seconds since the Unix epoch
        indicator_df['timestamp'] = pd.to_datetime(indicator_df['timestamp']).astype(int) // 10**9

        # Replace NaN and infinite values with 0
        indicator_df.fillna(value=0, inplace=True)

        # limit the DataFrame to the specified limit
#         indicator_df = indicator_df.tail(limit)

        end_ms = int(round(time.time() * 1000))
        indicator_time = end_ms - start_ms

        return {
            "symbol": symbol,
            "interval": interval,
            "kline_time": kline_time,
            "indicator_time": indicator_time,
            "proxy_server": proxy_server,
            "data": indicator_df.to_dict(orient='records')
        }
    except Exception as e:
        print(e)
        print(traceback.format_exc())
        raise HTTPException(
            status_code=500,
            detail=f"{str(e)}"
        )

"""On Balance Volume (OBV)"""
@app.get("/obv/{symbol}")
async def get_obv(
    symbol: str,
    interval: str = '1h',
    limit: Optional[int] = 2000,
    from_timestamp: Optional[int] = None,
):
    try:
        # Fetch kline data
        symbol = symbol.upper()

        # Process data
        df, kline_time, proxy_server = process_kline_data(symbol, interval, limit + 100, from_timestamp)

        start_ms = int(round(time.time() * 1000))

        # Calculate the On Balance Volume
        indicator = df.ta.obv()

        # Calculate the min
        indicator_df = indicator.reset_index()
        indicator_df.columns = [
            'timestamp',
            'value'
        ]

        # Convert the timestamp to seconds since the Unix epoch
        indicator_df['timestamp'] = pd.to_datetime(indicator_df['timestamp']).astype(int) // 10**9

        # Replace NaN and infinite values with 0
        indicator_df.fillna(value=0, inplace=True)

        # limit the DataFrame to the specified limit
#         indicator_df = indicator_df.tail(limit)

        end_ms = int(round(time.time() * 1000))
        indicator_time = end_ms - start_ms

        return {
            "symbol": symbol,
            "interval": interval,
            "kline_time": kline_time,
            "indicator_time": indicator_time,
            "proxy_server": proxy_server,
            "data": indicator_df.to_dict(orient='records')
        }
    except Exception as e:
        print(e)
        print(traceback.format_exc())
        raise HTTPException(
            status_code=500,
            detail=f"{str(e)}"
        )

"""Hilbert Transform - Instantaneous Trendline (HT_TRENDLINE)"""
@app.get("/ht_trendline/{symbol}")
async def get_ht_trendline(
    symbol: str,
    interval: str = '1h',
    limit: Optional[int] = 2000,
    from_timestamp: Optional[int] = None,
):
    try:
        # Fetch kline data
        symbol = symbol.upper()

        # Process data
        df, kline_time, proxy_server = process_kline_data(symbol, interval, limit + 100, from_timestamp)

        start_ms = int(round(time.time() * 1000))

        # Calculate the HT_TRENDLINE using pandas
        df['ht_trendline'] = df['close'].rolling(window=20).apply(lambda x: (x[-1] + x[-2]) / 2)

        # Convert HT_TRENDLINE to a DataFrame for consistent response formatting
        ht_trendline_df = df[['timestamp', 'ht_trendline']].reset_index(drop=True)
        ht_trendline_df.columns = ['timestamp', 'value']

        # Convert the timestamp to seconds since the Unix epoch
        ht_trendline_df['timestamp'] = pd.to_datetime(ht_trendline_df['timestamp']).astype(int) // 10**9

        # Replace NaN and infinite values with 0
        ht_trendline_df.fillna(value=0, inplace=True)

        end_ms = int(round(time.time() * 1000))
        indicator_time = end_ms - start_ms

        return {
            "symbol": symbol,
            "interval": interval,
            "kline_time": kline_time,
            "indicator_time": indicator_time,
            "proxy_server": proxy_server,
            "data": ht_trendline_df.to_dict(orient='records')
        }
    except Exception as e:
        print(e)
        print(traceback.format_exc())
        raise HTTPException(
            status_code=500,
            detail=f"{str(e)}"
        )

"""Ultimate Oscillator (UO)"""
@app.get("/uo/{symbol}")
async def get_uo(
    symbol: str,
    interval: str = '1h',
    limit: Optional[int] = 2000,
    from_timestamp: Optional[int] = None,
):
    try:
        # Fetch kline data
        symbol = symbol.upper()

        # Process data
        df, kline_time, proxy_server = process_kline_data(symbol, interval, limit + 100, from_timestamp)

        start_ms = int(round(time.time() * 1000))

        # Calculate the On Balance Volume
        indicator = df.ta.uo()

        # Calculate the min
        indicator_df = indicator.reset_index()
        indicator_df.columns = [
            'timestamp',
            'value'
        ]

        # Convert the timestamp to seconds since the Unix epoch
        indicator_df['timestamp'] = pd.to_datetime(indicator_df['timestamp']).astype(int) // 10**9

        # Replace NaN and infinite values with 0
        indicator_df.fillna(value=0, inplace=True)

        # limit the DataFrame to the specified limit
#         indicator_df = indicator_df.tail(limit)

        end_ms = int(round(time.time() * 1000))
        indicator_time = end_ms - start_ms

        return {
            "symbol": symbol,
            "interval": interval,
            "kline_time": kline_time,
            "indicator_time": indicator_time,
            "proxy_server": proxy_server,
            "data": indicator_df.to_dict(orient='records')
        }
    except Exception as e:
        print(e)
        print(traceback.format_exc())
        raise HTTPException(
            status_code=500,
            detail=f"{str(e)}"
        )

if __name__ == "__main__":
    uvicorn.run(app, host="0.0.0.0", port=9218, reload=True)